{-# LANGUAGE DeriveGeneric #-}
{-# LANGUAGE DuplicateRecordFields #-}
{-# LANGUAGE DeriveAnyClass #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE OverloadedStrings #-}

module Data.HexString
  ( HexString(..)
  , hexString
  , fromRawBytes
  , fromBinary
  , toBytes
  , fromText
  , toText
  ) where

import Codec.Borsh
import Control.Applicative (pure)

import Data.Aeson
import Data.Word (Word8)

import qualified Data.ByteString as BS
import qualified Data.ByteString.Base16 as BS16 (decode, decodeLenient, encode)
import qualified Data.ByteString.Lazy as BSL

import Data.Structured
import qualified Data.Text as T
import qualified Data.Text.Encoding as TE

import qualified Data.Binary as B (Binary, decode, encode)

import qualified GHC.Generics as GHC
import qualified Generics.SOP as SOP
import Text.Read

-- | Represents a Hex string. Guarantees that all characters it contains
--   are valid hex characters.
newtype HexString = HexString
  { hexBytes :: BS.ByteString
  } deriving stock (Eq, GHC.Generic)
    deriving anyclass (SOP.Generic, SOP.HasDatatypeInfo)
    deriving anyclass (Data.Structured.Show)
    deriving (BorshSize, ToBorsh, FromBorsh) via AsStruct HexString

instance Prelude.Show HexString where
  show = T.unpack . toText

instance Read HexString where
  readsPrec _ s = [((fromText . T.pack) s, "")]

instance FromJSON HexString where
  parseJSON = withText "HexString" $ pure . hexString . TE.encodeUtf8

instance ToJSON HexString where
  toJSON = Data.Aeson.String . toText

-- | Smart constructor which validates that all the text are actually
--   hexadecimal characters.
hexString :: BS.ByteString -> HexString
hexString bs =
  case BS16.decode bs of
    Right s -> HexString s
    Left e -> error e

-- | Converts a 'B.Binary' to a 'HexString' value
fromBinary :: B.Binary a => a -> HexString
fromBinary = HexString . BSL.toStrict . B.encode

-- | Converts a 'HexString' to a 'B.Binary' value
toBinary :: B.Binary a => HexString -> a
toBinary (HexString bs) = (B.decode . BSL.fromStrict) bs

-- | Reads a 'BS.ByteString' as raw bytes and converts to hex representation. We
--   cannot use the instance Binary of 'BS.ByteString' because it provides
--   a leading length, which is not what we want when dealing with raw bytes.
fromRawBytes :: BS.ByteString -> HexString
fromRawBytes = HexString

-- | Access to the raw bytes in a 'BS.ByteString' format.
toBytes :: HexString -> BS.ByteString
toBytes (HexString bs) = bs

-- | Reads a human-readable hex string into a `HexString`
fromText :: T.Text -> HexString
fromText = hexString . TE.encodeUtf8

-- | Access to a 'T.Text' representation of the 'HexString'
toText :: HexString -> T.Text
toText (HexString bs) = (TE.decodeUtf8 . BS16.encode) bs
